#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import io
import sys
import csv
import json
import uuid
import jinja2
import logging

try:
    import requests
except ImportError as e:
    pass

tpl_suffix = '-dstu2'

def rand_id():
    return str(uuid.uuid4())[-12:]


def ucfirst(string):
    return string[:1].upper() + string[1:]


def parse_int(string):
    try:
        return int(string)
    except ValueError:
        return None


def parse_float(string):
    try:
        return float(string)
    except ValueError:
        return None


def parse_bool(string):
    try:
        return bool(int(string))
    except ValueError:
        return None


def populate_demographics(data):
    data['pat_id'] = 'TIHM-pat-' + data['PtID']
    data['bday'] = {}
    data['bday'] = data['BD']
    data['name'] = {}
    data['name']['given'] = data['Name']
    data['name']['family'] = data['Name2']
    return data


def populate_mutations(data):
    obs = []
    obs.append({
        'id': "weight",
        'Weight': True,
        'code': "3141-9",
        'system': "http://lonic.org",
        'display': "Weight Measured",
        'quantity': parse_int(data['Weight']),
        'usystem': "http://unitsofmeasure.org",
        'ucum': 'lbs',
        'issued': data['WIssued'],
    })
    obs.append({
        'id': "bodytemperature",
        'code': "61008-9",  # body surface temp
        'system': "http://lonic.org",
        'display': "Body Temperature Measured",
        'quantity': parse_float(data['BodyTemperature']),
        'usystem': "http://unitsofmeasure.org",
        'ucum': 'Celsius',
        'issued': data['BTIssued'],
    })
    obs.append({
        'id': "roomtemperature",
        'code': "60832-3",
        'system': "http://lonic.org",
        'display': "Room Temperature Measured",
        'quantity': parse_float(data['RoomTemperature']),
        'usystem': "http://unitsofmeasure.org",
        'ucum': 'Celsius',
        'issued': data['RTIssued'],
    })
    obs.append({
        'id': "humidity",
        'code': "65643-9",
        'system': "http://lonic.org",
        'display': "Humidity Measured",
        'quantity': parse_int(data['Humidity']),
        'usystem': "http://unitsofmeasure.org",
        'ucum': '%',
        'issued': data['HIssued'],
    })
    obs.append({
        'id': "dispenser",
        'dispenser': True,
        'code': "TIHM-02",
        'system': "http://iot.surrey.ac.uk/tihm/models/fhir4tihm",
        'display': "Pill Dispenser",
        'dispensed': parse_bool(data['DSP']),
        'issued': data['DSPIssued'],
    })
    obs.append({
        'id': "motion",
        'motionsensor': True,
        'code': "TIHM-01",
        'system': "http://iot.surrey.ac.uk/tihm/models/fhir4tihm",
        'display': "Motion Sensor",
        'moved': parse_bool(data['MOTION']),
        'issued': data['MIssued'],
    })
    obs.append({
        'id': "pulseox",
        'pox': True,
        'code': "150456",
        'system': "http://rtmms.nist.gov",
        'display': "MDC_PULS_OXIM_SAT_02",
        'quantity': parse_int(data['POX']),
        'ucum': "%",
        'usystem': "http://rtmms.nist.gov",
        'issued': data['PIssued'],
    })
    obs.append({
        'id': "ecg",
        'ecg': True,
        'code': "131329",
        'system': "urn:oid:2.16.840.1.113883.6.24",
        'display': "MDC_ECG_ELEC_POTL_I",
        'data': data['ECG'],
        'usystem': "urn:oid:2.16.840.1.113883.6.24",
        'issued': data['ECGIssued'],
    })

    if len(obs) > 0:
        data['mutations'] = obs
    return data


def populate_labs(data):
    labs = []

    labval = data.get('BP')
    if labval:
        p1 = None
        p2 = None
        if '<' == labval[:1]:
            p2 = labval[1:]
        elif '>' == labval[:1]:
            p1 = labval[1:]
        elif '-' in labval:
            p1, p2 = labval.split('-')
        labs.append({
            'id': "blood pressure",
            'code': "55284-4",
            'system': "http://loinc.org",
            'display': "Blood pressure systolic & diastolic",
            'compon': True,
            'p1': p1,
            'codeL': "8480-6",
            'displayL': "Systolic blood pressure",
            'p2': p2,
            'codeH': "8462-4",
            'displayH': "Diastolic blood pressure",
            'string': labval,
            'ucum': "mm[Hg]",
            'usystem': "http://unitsofmeasure.org",
            'issued': data['BPIssued'],
        })

    labs = []
    labval = data.get('GPS')
    if labval:
        long, lat, alt = labval.split('--')

        labs.append({
            'id': "gps",
            'code': "56816-2",
            'system': "http://loinc.org",
            'display': "Patient Location",
            'gps': True,
            'longitude': long,
            'displayLong': "Longitude",
            'latitude': lat,
            'displayLat': "Latitude",
            'altitude': alt,
            'displayAlt': "Altitude",
            'string': labval,
            'issued': data['GPSIssued'],
        })

    data['labs'] = labs
    return data


def render(template, data, **args):
    d = data.copy()
    d.update(args)
    return template.render(rand_id=rand_id, **d)


def fhir_post_bundle(base_url, bundle):
    try:
        res = requests.post(base_url, headers={'Content-Type': 'application/json+fhir'}, data=bundle)
        res.raise_for_status()
    except Exception as e:
        logging.error("Failed; failing Bundle was:\n{}".format(bundle))
        raise e


def fhir_update(base_url, resource_type, resource):
    try:
        parsed = json.loads(resource)
        url = '{}/{}/{}'.format(base_url, resource_type, parsed['id'])
        res = requests.put(url, headers={'Content-Type': 'application/json+fhir'}, data=resource)
        res.raise_for_status()
    except Exception as e:
        logging.error("Failed; failing resource \"{}\" was:\n{}".format(resource_type, resource))
        raise e


if '__main__' == __name__:
    push_to = sys.argv[1] if len(sys.argv) > 1 else None
    bundle_per_patient = False

    # read CSV
    with io.open('fhir4tihm.csv', 'r') as csvfile:
        f = csv.reader(csvfile)
        head = None
        bundles = []
        resources = []

        # Jinja2
        tplenv = jinja2.Environment(loader=jinja2.PackageLoader(__name__, 'templates'))
        tpl_bundle = tplenv.get_template('Bundle{}.json'.format(tpl_suffix))
        tpl_patient = tplenv.get_template('patient{}.json'.format(tpl_suffix))
        tpl_observation = tplenv.get_template('observation{}.json'.format(tpl_suffix))

        # loop
        for row in f:
            if head is None:
                head = row

            # one row, one patient
            else:
                dic = dict(zip(head, row))
                demo = populate_demographics(dic)
                mutations = populate_mutations(dic)
                labs = populate_labs(dic)
                #gps = populate_gps(dic)


                resources.append(('Patient', render(tpl_patient, demo)))
                for obs in mutations.get('mutations', []): # for weight, temperatures, ECG, dispenser, pulse oximeter, motion, humidity
                    resources.append(('Observation', render(tpl_observation, demo, lab=obs)))

                for obs in labs.get('labs', []): # for Blood pressure and GPS
                    resources.append(('Observation', render(tpl_observation, demo, lab=obs)))

                bundle = tpl_bundle.render(bundle={'type': 'transaction', 'entries': [r[1] for r in resources]})
                bundles.append(bundle)

                if push_to is not None:
                    logging.debug("Pushing bundle for row {} to {}".format(row[0], push_to))
                    fhir_post_bundle(push_to, bundle)
                elif bundle_per_patient:
                    print(bundle)

        # create a master Bundle
        text_file = open('fhir4tihm.json', 'w+')
        if not bundle_per_patient:
            bundle = tpl_bundle.render(bundle={'type': 'transaction', 'entries': [r[1] for r in resources]})

        text_file.write(bundle)
        text_file.flush()
        text_file.close()

    #logging.debug('Done')
